# -*- coding: utf-8 -*-
import json
import os
import sys

import yaml
from clinner.command import command
from clinner.exceptions import ImproperlyConfigured
from clinner.run import Main

from health_check.settings import settings
from health_check.providers import Resource

OUTPUT_FORMAT_JSON = 'json'
OUTPUT_FORMAT_YAML = 'yaml'
OUTPUT_FORMAT_CHOICES = (OUTPUT_FORMAT_JSON, OUTPUT_FORMAT_YAML)


def _get_resource_filtered(resource_name, providers):
    """
    Get a resource by name and filter providers result with given list.
    
    :param resource_name: Resource name.
    :param providers: Providers list.
    :return: All providers info of given result.
    """
    resource = Resource(resource_name)

    data = resource()

    if providers and data:
        data = {k: v for k, v in data.items() if k in providers}

    return data


def _print_resource(data, output_format):
    """
    Prints resource info using given format.
    
    :param data: Resource info.
    :param output_format: Output format.
    """
    if output_format == OUTPUT_FORMAT_JSON:
        output = json.dumps(data)
    else:
        output = yaml.safe_dump(data, default_flow_style=False)

    print(output)


@command(args=((('-e', '--error-on-fail'), {'help': 'Returns error code if a health provider fails',
                                            'action': 'store_true', 'default': False}),
               (('-f', '--output-format'), {
                   'help': 'Output format', 'choices': OUTPUT_FORMAT_CHOICES,
                   'default': OUTPUT_FORMAT_YAML
               }),
               (('provider',), {'help': 'Providers to be used for checking', 'nargs': '*'})),
         parser_opts={'help': 'Run health providers'})
def health(*args, **kwargs):
    data = _get_resource_filtered('health', kwargs.get('provider'))

    _print_resource(data, kwargs['output_format'])

    if kwargs['error_on_fail'] and not all([value for provider in data.values() for value in provider.values()]):
        sys.exit(1)


@command(args=((('-f', '--output-format'), {'help': 'Output format', 'choices': OUTPUT_FORMAT_CHOICES,
                                            'default': OUTPUT_FORMAT_YAML}),
               (('provider',), {'help': 'Providers to be used for checking', 'nargs': '*'})),
         parser_opts={'help': 'Run stats providers'})
def stats(*args, **kwargs):
    data = _get_resource_filtered('stats', kwargs.get('provider'))

    _print_resource(data, kwargs['output_format'])


class HealthCheckMain(Main):
    """
    Command for running producer, consumer and scheduler processes along with some other utilities. This acts as a
    Command Line Interface for Task dispatcher.
    """
    description = 'Health check command that provides a common entry point for running the different checks'

    def inject_app_settings(self):
        if self.settings:
            os.environ['HEALTH_CHECK_SETTINGS'] = self.settings

        if 'HEALTH_CHECK_SETTINGS' not in os.environ:
            raise ImproperlyConfigured('Settings not defined')

        settings.build_from_module(os.environ['HEALTH_CHECK_SETTINGS'])



# -*- coding: utf-8 -*-
"""
Built-in Django check providers.
"""
import datetime

from django.core.cache import caches as django_caches
from django.db import connections

from health_check.settings import settings


def databases(*args, **kwargs):
    """
    Check database health_check.

    :return: Status of each database.
    """
    status = {}
    for connection in connections.all():
        try:
            connection.connect()
            status[connection.alias] = connection.is_usable()
        except:
            status[connection.alias] = False

    return status


def caches(*args, **kwargs):
    """
    Check caches health_check.

    :return: Status of each cache.
    """
    caches_aliases = settings.caches.keys()
    value = datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')

    status = {}
    for alias in caches_aliases:
        try:
            cache = django_caches[alias]
            cache.set('health_check_test_cache', value)
            status[alias] = True
        except:
            status[alias] = False

    return status



# -*- coding: utf-8 -*-
import importlib

from health_check.settings import settings

__all__ = ['Provider', 'Resource']


class Provider:
    """
    Wrapper that handles a check provider, giving easy access to them with dynamic imports from string.
    """

    def __init__(self, name, provider, args, kwargs):
        """
        Create a provider.

        :param name: Provider name.
        :param provider: Provider function or string.
        :param args: Provider args.
        :param kwargs: Provider kwargs.
        """
        self.name = name

        if isinstance(provider, str):
            provider_module, provider_func = provider.rsplit('.', 1)
            module = importlib.import_module(provider_module)
            self.provider = getattr(module, provider_func, None)
            if self.provider is None:
                raise ValueError('Provider not found: %s' % (provider,))
        else:
            self.provider = provider

        self.args = args or ()
        self.kwargs = kwargs or {}

    def __call__(self, *args, **kwargs):
        """
        Return provider results.

        :param args: Provider args.
        :param kwargs: Provider kwargs
        :return: Results after evaluates provider.
        """
        if not args:
            args = self.args

        if not kwargs:
            kwargs = self.kwargs

        return self.provider(*args, **kwargs)


class Resource:
    """
    Wrapper that handles a whole resource with its providers.
    """

    def __init__(self, name):
        """
        Create a resource and all its providers.

        :param name: Resource name
        """
        self.name = name
        try:
            self.providers = {name: Provider(name=name, provider=provider, args=args, kwargs=kwargs)
                              for name, provider, args, kwargs in settings.providers[self.name]}
        except KeyError:
            raise ValueError("Resource doesn't exists: %s" % (self.name,))

    def __call__(self, *args, **kwargs):
        """
        Return results from all providers of this resource.

        :return: Results after evaluate each provider.
        """
        return {name: provider() for name, provider in self.providers.items()}



# -*- coding: utf-8 -*-
"""
Built-in Django check providers.
"""

from django.db import connections


def databases(*args, **kwargs):
    """
    Retrieve the stats data of each database.

    :return: Stats data of each database.
    """
    stats = {}
    for connection in connections.all():
        try:
            connection.connect()
        except:
            is_usable = False
        else:
            is_usable = connection.is_usable()
        finally:
            stats[connection.alias] = {
                'vendor': connection.vendor,
                'is_usable': is_usable,
                'allow_thread_sharing': connection.allow_thread_sharing,
                'autocommit': connection.autocommit,
                'commit_on_exit': connection.commit_on_exit,
                'in_atomic_block': connection.in_atomic_block,
                'settings': {k: v for k, v in connection.settings_dict.items() if k not in ('USER', 'PASSWORD')}
            }

    return stats



# -*- coding: utf-8 -*-
"""
Django HealthCheck application.
"""
__version__ = '3.1.2'
__license__ = 'GPLv3'

__author__ = 'José Antonio Perdiguero López'
__email__ = 'perdy.hh@gmail.com'

__url__ = 'https://github.com/PeRDy/health-check'
__description__ = 'Health Check is an application that provides an API to check the health of some parts and some ' \
                  'utilities like ping requests. This application can works as standalone or included in a Django ' \
                  'project. '

default_app_config = 'health_check.apps.HealthCheck'



# -*- coding: utf-8 -*-
"""
Built-in Celery check providers.
"""

from health_check.utils import FakeChecker

try:
    from celery.task import control

    celery_inspect = control.inspect()
except ImportError:
    celery_inspect = FakeChecker()


def celery(workers, *args, **kwargs):
    """
    Check if given celery workers are running.

    :param workers: List of workers to be checked.
    :return: Status of each worker.
    """
    try:
        ping_response = celery_inspect.ping() or {}
        active_workers = ping_response.keys()
        workers_status = {w: w in active_workers for w in workers}
    except (AttributeError, OSError):
        workers_status = None

    return workers_status



# -*- coding: utf-8 -*-
"""
URLs.
"""
from itertools import chain

from django.conf.urls import url

from health_check.settings import settings
from health_check.api.views import RootAPIView, ProviderAPIView, ResourceAPIView


def get_provider_urls(resource, providers):
    # Resource root view
    urls = [url(r'^{}/?$'.format(resource), ResourceAPIView.as_view(resource=resource),
                name='api_{}_root'.format(resource))]

    # Resource providers views
    urls += [url(r'^{}/{}/?$'.format(resource, n),
                 ProviderAPIView.as_view(provider=p, provider_name=n, provider_args=args, provider_kwargs=kwargs),
                 name='api_{}_{}'.format(resource, n))
             for n, p, args, kwargs in providers]

    return urls


providers_urls = chain.from_iterable([get_provider_urls(r, p) for r, p in settings.providers.items()])

urlpatterns = list(providers_urls)

urlpatterns += [
    url(r'^$', RootAPIView.as_view())
]



# -*- coding: utf-8 -*-
from clinner.run import DjangoCommand
from health_check.commands import HealthCheckMain


class Command(DjangoCommand):
    """
    Wrapper that makes a Django command from TaskDispatcherCommand.
    """
    main_class = HealthCheckMain



# -*- coding: utf-8 -*-
"""
Settings.
"""
import os
from functools import partial
from functools import update_wrapper
from importlib import import_module

__all__ = ['settings']


class reset:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        update_wrapper(self, func)

    def __get__(self, instance, owner=None):
        self.instance = instance
        return partial(self, instance)

    def __call__(self, *args, **kwargs):
        self.instance.reset_default()
        self.func(*args, **kwargs)


class Settings:
    debug = False
    base_dir = None
    installed_apps = ()
    caches = {}
    celery_workers = ()
    providers = {'health': (), 'stats': ()}

    def __init__(self):
        self.reset_default()
        module_path = os.environ.get('HEALTH_CHECK_SETTINGS')
        if module_path:
            self.build_from_module(module_path)

    def reset_default(self):
        """
        Reset settings to default values.
        """
        self.debug = False
        self.base_dir = None
        self.installed_apps = ()
        self.caches = {}
        self.celery_workers = ()
        self.providers = {
            'health': (
                ('ping', 'health_check.providers.health.ping', None, None),
            ),
            'stats': (
                ('code', 'health_check.providers.stats.code', None, None),
            )
        }

    @staticmethod
    def import_settings(path):
        """
        Import a settings module that can be a module or an object.
        To import a module, his full path should be specified: *package.settings*.
        To import an object, his full path and object name should be specified: *project.settings:SettingsObject*.

        :param path: Settings full path.
        :return: Settings module or object.
        """
        try:
            try:
                m, c = path.rsplit(':', 1)
                module = import_module(m)
                s = getattr(module, c)
            except ValueError:
                s = import_module(path)
        except ImportError:
            raise ImportError("Settings not found '{}'".format(path))

        return s

    @staticmethod
    def get(s, key, default=None):
        """
        Get a settings value from key. Try to get key as it is, transforming to lower case and to upper case. If not
        found, a default value will be returned.

        :param s: Settings module or object.
        :param key: Settings key.
        :param default: Default value.
        :return: Settings value. Default value if key is not found.
        """
        return getattr(s, key, getattr(s, key.lower(), getattr(s, key.upper(), default)))

    def _add_celery_providers(self):
        self.providers['health'] += (
            ('celery', 'health_check.providers.celery.health.celery', None, {'workers': self.celery_workers}),
        )
        self.providers['stats'] += (
            ('celery', 'health_check.providers.celery.stats.celery', None, {'workers': self.celery_workers}),
        )

    @reset
    def build_from_module(self, module=None):
        if isinstance(module, str):
            module = self.import_settings(module)

        # Project path defined in settings
        self.base_dir = self.get(module, 'base_dir')

        # Celery application
        self.celery_workers = self.get(module, 'health_check_celery_workers', ())

        # Mapping of resources to list of providers
        # Each provider is a tuple of application name, provider, args, kwargs
        self.providers = self.get(module, 'health_check_providers', {
            'health': (
                ('ping', 'health_check.providers.health.ping', None, None),
            ),
            'stats': (
                ('code', 'health_check.providers.stats.code', None, None),
            )
        })

        # If there is celery workers then add celery providers
        if self.celery_workers:
            self._add_celery_providers()

    @reset
    def build_from_django(self):
        from django.conf import settings as django_settings

        # Django debug settings.
        self.debug = self.get(django_settings, 'DEBUG', False)

        # Django installed apps
        self.installed_apps = self.get(django_settings, 'INSTALLED_APPS', [])
        self.caches = self.get(django_settings, 'CACHES', {})

        # Project path defined in Django settings
        self.base_dir = self.get(django_settings, 'BASE_DIR')

        # Celery application
        self.celery_workers = self.get(django_settings, 'HEALTH_CHECK_CELERY_WORKERS', ())

        # Mapping of resources to list of providers
        # Each provider is a tuple of application name, provider, args, kwargs
        self.providers = self.get(django_settings, 'HEALTH_CHECK_PROVIDERS', {
            'health': (
                ('ping', 'health_check.providers.health.ping', None, None),
                ('databases', 'health_check.providers.django.health.databases', None, None),
                ('caches', 'health_check.providers.django.health.caches', None, None),
            ),
            'stats': (
                ('databases', 'health_check.providers.django.stats.databases', None, None),
                ('code', 'health_check.providers.stats.code', None, None),
            )
        })

        # If there is celery workers then add celery providers
        if self.celery_workers:
            self._add_celery_providers()


settings = Settings()



# -*- coding: utf-8 -*-
"""
Built-in check providers.
"""


def ping(*args, **kwargs):
    """
    Check if current application is running.

    :return: Pong response.
    """

    return {'pong': True}
